/**
 * Copyright (c) 2020 Bubble, Inc.  All rights reserved.
 * For personal (non-commercial) use, see license: https://getbubblenow.com/bubble-license/
 */
extern crate clap;
extern crate bcrypt;
extern crate system_shutdown;

use clap::{Arg, ArgMatches, App as ClapApp};
use std::env;
use std::fmt::Error;
use std::fs;
use std::fs::File;
use std::os::unix::fs::PermissionsExt;
use std::io::{Write, Read, BufWriter};
use std::process::{Command, Stdio};
use std::path::Path;
use std::time::{SystemTime, UNIX_EPOCH};
use actix_web::{web, middleware, App as ActixApp, HttpResponse, HttpServer, Responder, Error as ActixError};
use actix_web::web::Bytes;
use actix_web::dev::ServiceRequest;
use actix_web_httpauth::extractors::basic::BasicAuth;
use actix_web_httpauth::middleware::HttpAuthentication;
use reqwest;
use reqwest::StatusCode;
use probes;
use tempfile;
use tempfile::TempDir;
use nix::unistd;
use nix::unistd::{Uid, Gid};
use users::{get_user_by_name, get_group_by_name};


const ADMIN_USERNAME : &'static str = "bubble_node_admin";
const BUBBLE_NODEMANAGER : &'static str = "bubble-nodemanager";
const ROOT_DIR_PREFIX : &'static str = "root_dir/";
static mut CONFIG : Option<ArgMatches> = None;

const BUBBLE_PATCH : &'static str = "bubble";
const MITMPROXY_PATCH : &'static str = "mitmproxy";
const ROOT_PATCH : &'static str = "root";

const VALID_PATCHES : [&str; 3] = [BUBBLE_PATCH, MITMPROXY_PATCH, ROOT_PATCH];

const ENV_DEFAULT_PASSWORD_FILE : &'static str = "BUBBLE_NODEMANAGER_PASSWORD_FILE";
const DEFAULT_PASSWORD_FILE : &'static str = "/home/bubble/.nodemanager_pass";

const ACTION_START   : &'static str = "start";
const ACTION_STOP    : &'static str = "stop";
const ACTION_RESTART : &'static str = "restart";
const ACTION_RELOAD  : &'static str = "reload";
const ACTION_STATUS  : &'static str = "status";

const ACTIONS_SUPERVISOR: &'static [&'static str] = &[ACTION_RELOAD, ACTION_STATUS];
const ACTIONS_STANDARD : &'static [&'static str] = &[ACTION_START, ACTION_STOP, ACTION_RESTART, ACTION_STATUS];
const ACTIONS_NODEMANAGER : &'static [&'static str] = &[ACTION_START, ACTION_RESTART, ACTION_STATUS];

const CMD_SUPERVISORCTL : &'static str = "supervisorctl";
const CMD_SERVICE : &'static str = "service";

struct BubbleService {
    name : &'static str,
    service : &'static str,
    command : &'static str,
    actions : &'static [&'static str]
}

static SVC_SUPERVISOR : BubbleService = BubbleService {
    name : "supervisor",
    service : "",
    command : CMD_SUPERVISORCTL,
    actions : ACTIONS_SUPERVISOR
};

static SVC_BUBBLE : BubbleService = BubbleService {
    name : "bubble",
    service : "bubble",
    command : CMD_SUPERVISORCTL,
    actions : ACTIONS_STANDARD
};

static SVC_MITMPROXY : BubbleService = BubbleService {
    name : "mitm",
    service : "mitmdump",
    command : CMD_SUPERVISORCTL,
    actions : ACTIONS_STANDARD
};

static SVC_NGINX : BubbleService = BubbleService {
    name : "nginx",
    service : "nginx",
    command : CMD_SERVICE,
    actions : ACTIONS_STANDARD
};

static SVC_POSTGRESQL : BubbleService = BubbleService {
    name : "postgresql",
    service : "postgresql",
    command : CMD_SERVICE,
    actions : ACTIONS_STANDARD
};

static SVC_NODEMANAGER : BubbleService = BubbleService {
    name : "nodemanager",
    service : "nodemanager",
    command : CMD_SUPERVISORCTL,
    actions : ACTIONS_NODEMANAGER
};

static BUBBLE_SERVICES : [&BubbleService; 6] = [
    &SVC_SUPERVISOR, &SVC_BUBBLE, &SVC_MITMPROXY, &SVC_NGINX, &SVC_POSTGRESQL, &SVC_NODEMANAGER
];

fn patch_type (component : &String) -> Result<String, Error> {
    if VALID_PATCHES.iter().any(|c| c.eq(component)) {
        Ok(String::from(component))
    } else {
        Err(Error)
    }
}

fn find_service (name : &String) -> Result<&BubbleService, Error> {
    let service = BUBBLE_SERVICES.iter().find(|s| (*s).name.eq(name));
    if service.is_none() {
        Err(Error)
    } else {
        Ok(service.unwrap())
    }
}

async fn validator(
    req: ServiceRequest,
    credentials: BasicAuth,
) -> Result<ServiceRequest, ActixError> {
    if !credentials.user_id().eq(ADMIN_USERNAME) {
        Err(ActixError::from(HttpResponse::NotFound()))
    } else {
        let pass = credentials.password();
        if pass.is_none() {
            Err(ActixError::from(HttpResponse::BadRequest()))
        } else {
            let password = pass.unwrap();
            if !bcrypt::verify(&password.as_ref(), read_hashed_password().as_str()).unwrap() {
                Err(ActixError::from(HttpResponse::Forbidden()))
            } else {
                Ok(req)
            }
        }
    }
}

fn read_hashed_password() -> String {
    unsafe {
        let password_file;
        let password_file_path = CONFIG.as_ref().unwrap().value_of("password_file");
        if password_file_path.is_none() {
            let password_file_env = env::var(ENV_DEFAULT_PASSWORD_FILE);
            if password_file_env.is_err() {
                println!("read_hashed_password: no command-line option provided, and env var not found, using default: {}", DEFAULT_PASSWORD_FILE);
                password_file = String::from(DEFAULT_PASSWORD_FILE);
            } else {
                password_file = password_file_env.unwrap();
            }
        } else {
            password_file = String::from(CONFIG.as_ref().unwrap().value_of("password_file").unwrap());
        }
        let result = fs::read_to_string(&password_file);
        let contents = result.expect(format!("Error reading password file: {}", password_file).as_str());
        contents.trim().to_string()
    }
}

async fn ping() -> impl Responder {
    HttpResponse::Ok().body("PONG!\n")
}

async fn get_stats(path: web::Path<(String,)>) -> impl Responder {
    let stat_name = &path.0;
    if stat_name.eq("uptime") {
        let load = probes::load::read();
        if load.is_ok() {
            let load_avg = load.unwrap();
            HttpResponse::Ok().body(format!("{{ \"one\": {}, \"five\": {}, \"fifteen\": {} }}\n", load_avg.one, load_avg.five, load_avg.fifteen))
        } else {
            HttpResponse::InternalServerError().body(format!("get_stats: error reading load stats"))
        }

    } else if stat_name.eq("mem") {
        let memory = probes::memory::read();
        if memory.is_ok() {
            let mem = memory.unwrap();
            HttpResponse::Ok().body(format!("{{ \"total\": {}, \"used\": {}, \"free\": {} }}\n", mem.total(), mem.used(), mem.free()))
        } else {
            HttpResponse::InternalServerError().body(format!("get_stats: error reading memory stats"))
        }

    } else if stat_name.eq("net") {
        let network = probes::network::read();
        if network.is_ok() {
            let usage = network.unwrap();
            let mut usage_string = String::from("[\n");
            let mut first = true;
            for (net_if, net_stats) in usage.interfaces.iter() {
                if !first {
                    usage_string.push_str(",\n");
                } else {
                    first = false;
                }
                usage_string.push_str(format!("  {{ \"interface\": \"{}\", \"tx\": {}, \"rx\": {} }}", &net_if, &net_stats.transmitted, &net_stats.received).as_str());
            }
            usage_string.push_str("\n]\n");
            HttpResponse::Ok().body(usage_string)
        } else {
            HttpResponse::InternalServerError().body(format!("get_stats: error reading network traffic stats"))
        }

    } else if stat_name.eq("disk") {
        let disk = probes::disk_usage::read();
        if disk.is_ok() {
            let usage = disk.unwrap();
            let mut usage_string = String::from("[\n");
            let mut first = true;
            for u in usage.iter() {
                if !first {
                    usage_string.push_str(",\n");
                } else {
                    first = false;
                }
                usage_string.push_str(format!("  {{ \"mount\": \"{}\", \"pct_used\": {} }}", &u.mountpoint, &u.used_percentage).as_str());
            }
            usage_string.push_str("\n]\n");
            HttpResponse::Ok().body(usage_string)
        } else {
            HttpResponse::InternalServerError().body(format!("get_stats: error reading disk usage"))
        }

    } else {
        HttpResponse::NotFound().body(format!("get_stats: stat not found: {}\n", stat_name))
    }
}

fn component_path (component : &String, path : &str) -> Result<String, Error> {
    if component.eq(ROOT_PATCH) {
        if !path.starts_with(ROOT_DIR_PREFIX) {
            return Err(Error)
        }
        let sub_path = &path[ROOT_DIR_PREFIX.len()..];
        return Ok(format!("/{}", sub_path));

    } else if component.eq(BUBBLE_PATCH) {
        return Ok(format!("/home/bubble/{}", path));

    } else if component.eq(MITMPROXY_PATCH) {
        return Ok(format!("/home/mitmproxy/{}", path));
    }
    Err(Error)
}

fn unzip_and_apply_patch(component : &String, zipfile: File) -> Result<TempDir, Error> {
    let user = get_user_by_name(component);
    if user.is_none() {
        println!("{}: user not found", component);
        return Err(Error);
    }
    let uid = user.unwrap().uid();

    let group = get_group_by_name(component);
    if group.is_none() {
        println!("{}: group not found", component);
        return Err(Error);
    }
    let gid = group.unwrap().gid();

    let tempdir = tempfile::tempdir().unwrap();
    let mut zip = zip::ZipArchive::new(&zipfile).unwrap();
    for i in 0..zip.len() {
        let mut file = zip.by_index(i).unwrap();
        let path = component_path(component, file.name()).unwrap();
        if file.is_dir() {
            continue;
        }

        let out_path;
        let mut temp_path = None;
        if path.ends_with(BUBBLE_NODEMANAGER) {
            let now = SystemTime::now().duration_since(UNIX_EPOCH).unwrap().as_nanos();
            temp_path = Some(format!("{}.tmp{}", path, now));
            out_path = Path::new(temp_path.as_ref().unwrap().as_str());
        } else {
            out_path = Path::new(path.as_str());
        }

        fs::create_dir_all(out_path.parent().unwrap()).unwrap();
        println!("{}: zipfile = {}, path={}", component, file.name(), path);
        let mut buffer = [0 as u8; 1024];
        let new_file = !out_path.exists();
        let dest_file = File::create(out_path).unwrap();
        if new_file {
            println!("{}: setting ownership on path={} to owner={}/{}, group={}/{}", component, path, component, uid, component, gid);
            unistd::chown(out_path, Some(Uid::from_raw(uid)), Some(Gid::from_raw(gid))).unwrap();
        }
        let mut writer = BufWriter::new(&dest_file);
        loop {
            let bytes_read = file.read(& mut buffer).unwrap();
            if bytes_read == 0 {
                break;
            }
            let mut total_written = 0 as usize;
            loop {
                let bytes_written = writer.write(&buffer[total_written..bytes_read]).unwrap();
                total_written += bytes_written;
                if total_written == bytes_read {
                    break;
                } else if total_written > bytes_read {
                    return Err(Error);
                }
            }
            if total_written != bytes_read {
                return Err(Error);
            }
        }
        writer.flush().unwrap();
        if temp_path.is_some() {
            let real_meta = fs::metadata(path.as_str()).unwrap();
            let real_perms = real_meta.permissions();
            let src_file = File::open(out_path).unwrap();
            src_file.set_permissions(real_perms).unwrap();
            src_file.sync_all().unwrap();
            println!("temp_path={:?} : set src_perms={:?} for real_path={}", out_path, src_file.metadata().unwrap().permissions().mode(), path.as_str());
            fs::rename(out_path, Path::new(path.as_str())).unwrap();
        }
    }
    Ok(tempdir)
}

async fn patch_cmd(path: web::Path<(String,)>, data: Bytes) -> impl Responder {
    // validate component is one of: root, bubble, mitmproxy
    let component = patch_type(&path.0).unwrap();
    let url = std::str::from_utf8(data.as_ref()).unwrap();
    println!("patch_cmd: requesting patch: {}", url);
    match reqwest::get(url).await {
        Ok(mut response) => {
            match response.status() {
                StatusCode::OK => {
                    let tmpfile: File = tempfile::tempfile().unwrap();
                    {
                        let mut writer = BufWriter::new(&tmpfile);
                        loop {
                            let raw_chunk = response.chunk().await;
                            match raw_chunk {
                                Ok(chunk) => {
                                    match chunk {
                                        Some(data) => { writer.write(&data).unwrap(); },
                                        None => break
                                    }
                                },
                                Err(error) => return HttpResponse::InternalServerError().body(format!("patch({}): url={} chunk error {}", component, url, error))
                            }
                        }
                        if tmpfile.sync_all().is_err() {
                            return HttpResponse::InternalServerError().body(format!("patch({}): url={} error syncing file", component, url));
                        }
                    }
                    unzip_and_apply_patch(&component, tmpfile).unwrap();
                    HttpResponse::Ok().body(format!("patch({}): response = {:?}", component, response))
                },
                _ => HttpResponse::InternalServerError().body(format!("patch({}): url={}, http status code: {}", component, url, response.status().as_u16()))
            }
        },
        Err(error) => HttpResponse::InternalServerError().body(format!("patch: error: {}", error))
    }
}

async fn service_cmd(path: web::Path<(String, String)>) -> impl Responder {
    let service = &path.0;
    let action = &path.1;
    let bubble_service = find_service(service).unwrap();
    if !bubble_service.actions.iter().any(|a| a.eq(action)) {
        return HttpResponse::NotFound().body(format!("service_cmd: action not found: {}", action));
    }
    let cmd_line;
    if bubble_service.command.eq(CMD_SERVICE) {
        cmd_line = format!("{} {} {}", bubble_service.command, bubble_service.service, action);
    } else {
        cmd_line = format!("{} {} {}", bubble_service.command, action, bubble_service.service);
    }
    println!("service_cmd: running {}", &cmd_line);
    let result = Command::new("/bin/bash")
        .stdin(Stdio::null())
        .arg("-c")
        .arg(&cmd_line)
        .output();
    if result.is_ok() {
        let res = result.unwrap();
        if res.status.success() {
            if res.stdout.len() == 0 {
                HttpResponse::Ok().body("OK\n")
            } else {
                HttpResponse::Ok().body(res.stdout)
            }
        } else {
            HttpResponse::InternalServerError().body(res.stderr)
        }
    } else {
        HttpResponse::InternalServerError().body(format!("{:?}", result.err().unwrap()))
    }
}

async fn post_cmd(path: web::Path<(String,)>) -> impl Responder {
    let cmd = &path.0;
    if cmd.eq("reboot") {
        match system_shutdown::reboot() {
            Ok(_) => HttpResponse::Ok().body("rebooting!"),
            Err(error) => HttpResponse::InternalServerError().body(format!("post_cmd: error rebooting: {}", error))
        }
    } else {
        HttpResponse::NotFound().body(format!("post_cmd: unknown command: {}\n", cmd))
    }
}

#[actix_rt::main]
async fn main() -> std::io::Result<()> {
    let matches = ClapApp::new("bubble-nodemanager")
        .version("0.1.0")
        .author("Jonathan Cobb <jonathan@getbubblenow.com>")
        .about("Low-level management services for Bubble nodes")
        .arg(Arg::with_name("password_file")
            .short("w")
            .long("PASSWORD_FILE")
            .value_name("FILE")
            .help("Path to password file containing the bcrypt-encrypted password")
            .takes_value(true))
        .arg(Arg::with_name("port")
            .short("p")
            .long("port")
            .value_name("port")
            .help("Port to listen on")
            .default_value("7800")
            .takes_value(true))
        .get_matches();

    unsafe {
        CONFIG = Some(matches);
        let config = CONFIG.as_ref().unwrap();
        let port = config.value_of("port").unwrap();
        let bind_addr = format!("127.0.0.1:{}", port);
        println!("Listening on port {}", port);

        HttpServer::new(move || {
            let auth = HttpAuthentication::basic(validator);
            ActixApp::new()
                .wrap(middleware::Logger::default())
                .wrap(auth)
                .service(
                    web::scope("/")
                        .route("ping", web::get().to(ping))
                        .route("stats/{type}", web::get().to(get_stats))
                        .route("patch/{file}", web::post().to(patch_cmd))
                        .route("service/{svc}/{action}", web::post().to(service_cmd))
                        .route("cmd/{cmd}", web::post().to(post_cmd))
                )
        })
            .workers(4)
            .bind(&bind_addr)?
            .run()
            .await
    }
}
